package apply

import (
	"encoding/json"
	"fmt"
	"io"
	"log"
	"os"

	kubepolicy "github.com/nirmata/kyverno/pkg/apis/policy/v1alpha1"
	"github.com/nirmata/kyverno/pkg/engine"
	"github.com/spf13/cobra"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	unstructured "k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	schema "k8s.io/apimachinery/pkg/runtime/schema"
	yaml "k8s.io/apimachinery/pkg/util/yaml"
	memory "k8s.io/client-go/discovery/cached/memory"
	dynamic "k8s.io/client-go/dynamic"
	kubernetes "k8s.io/client-go/kubernetes"
	"k8s.io/client-go/kubernetes/scheme"
	"k8s.io/client-go/restmapper"
)

const applyExample = `  # Apply a policy to the resource.
  kyverno apply @policy.yaml @resource.yaml
  kyverno apply @policy.yaml @resourceDir/
  kyverno apply @policy.yaml @resource.yaml --kubeconfig=$PATH_TO_KUBECONFIG_FILE`

// NewCmdApply returns the apply command for kyverno
func NewCmdApply(in io.Reader, out, errout io.Writer) *cobra.Command {
	var kubeconfig string
	cmd := &cobra.Command{
		Use:     "apply",
		Short:   "Apply policy on the resource(s)",
		Example: applyExample,
		Run: func(cmd *cobra.Command, args []string) {
			policy, resources := complete(kubeconfig, args)
			output := applyPolicy(policy, resources)
			fmt.Printf("%v\n", output)
		},
	}

	cmd.Flags().StringVar(&kubeconfig, "kubeconfig", "", "path to kubeconfig file")
	return cmd
}

func complete(kubeconfig string, args []string) (*kubepolicy.Policy, []*resourceInfo) {

	policyDir, resourceDir, err := validateDir(args)
	if err != nil {
		fmt.Printf("Failed to parse file path, err: %v\n", err)
		os.Exit(1)
	}

	// extract policy
	policy, err := extractPolicy(policyDir)
	if err != nil {
		log.Printf("failed to extract policy: %v", err)
		os.Exit(1)
	}

	// extract rawResource
	resources, err := extractResource(resourceDir, kubeconfig)
	if err != nil {
		log.Printf("failed to parse resource: %v", err)
		os.Exit(1)
	}

	return policy, resources
}

func applyPolicy(policy *kubepolicy.Policy, resources []*resourceInfo) (output string) {
	for _, resource := range resources {
		patchedDocument, err := applyPolicyOnRaw(policy, resource.rawResource, resource.gvk)
		if err != nil {
			fmt.Printf("Error applying policy on resource %s, err: %v\n", resource.gvk.Kind, err)
			continue
		}

		out, err := prettyPrint(patchedDocument)
		if err != nil {
			fmt.Printf("JSON parse error: %v\n", err)
			continue
		}

		output = output + fmt.Sprintf("---\n%s", string(out))
	}
	return
}

func applyPolicyOnRaw(policy *kubepolicy.Policy, rawResource []byte, gvk *metav1.GroupVersionKind) ([]byte, error) {
	_, patchedDocument := engine.Mutate(*policy, rawResource, *gvk)

	if err := engine.Validate(*policy, patchedDocument, *gvk); err != nil {
		return nil, err
	}
	return patchedDocument, nil
}

func extractPolicy(fileDir string) (*kubepolicy.Policy, error) {
	policy := &kubepolicy.Policy{}

	file, err := loadFile(fileDir)
	if err != nil {
		return nil, fmt.Errorf("failed to load file: %v", err)
	}

	policyBytes, err := yaml.ToJSON(file)
	if err != nil {
		return nil, err
	}

	if err := json.Unmarshal(policyBytes, policy); err != nil {
		return nil, fmt.Errorf("failed to decode policy %s, err: %v", policy.Name, err)
	}

	if policy.TypeMeta.Kind != "Policy" {
		return nil, fmt.Errorf("failed to parse policy")
	}

	return policy, nil
}

type resourceInfo struct {
	rawResource []byte
	gvk         *metav1.GroupVersionKind
}

func extractResource(fileDir, kubeconfig string) ([]*resourceInfo, error) {
	var files []string
	var resources []*resourceInfo

	// check if applied on multiple resources
	isDir, err := isDir(fileDir)
	if err != nil {
		return nil, err
	}

	if isDir {
		files, err = ScanDir(fileDir)
		if err != nil {
			return nil, err
		}
	} else {
		files = []string{fileDir}
	}

	for _, dir := range files {
		data, err := loadFile(dir)
		if err != nil {
			fmt.Printf("Warning: errpr while loading file: %v\n", err)
			continue
		}

		decode := scheme.Codecs.UniversalDeserializer().Decode
		obj, gvk, err := decode([]byte(data), nil, nil)
		if err != nil {
			fmt.Printf("Warning: error while decoding YAML object, err: %s\n", err)
			continue
		}

		actualObj, err := convertToActualObject(kubeconfig, gvk, obj)
		if err != nil {
			fmt.Printf("Warning: error while converting resource %s to actual k8s object, err: %v\n", gvk.Kind, err)
			fmt.Printf("Apply policy on raw resource.\n")
		}

		raw, err := json.Marshal(actualObj)
		if err != nil {
			fmt.Printf("Warning: error while marshalling manifest, err: %v\n", err)
			continue
		}

		gvkInfo := &metav1.GroupVersionKind{Group: gvk.Group, Version: gvk.Version, Kind: gvk.Kind}
		resources = append(resources, &resourceInfo{rawResource: raw, gvk: gvkInfo})
	}

	return resources, err
}

func convertToActualObject(kubeconfig string, gvk *schema.GroupVersionKind, obj runtime.Object) (interface{}, error) {
	clientConfig, err := createClientConfig(kubeconfig)
	if err != nil {
		return obj, err
	}

	dynamicClient, err := dynamic.NewForConfig(clientConfig)
	if err != nil {
		return obj, err
	}

	kclient, err := kubernetes.NewForConfig(clientConfig)
	if err != nil {
		return obj, err
	}

	asUnstructured := &unstructured.Unstructured{}
	if err := scheme.Scheme.Convert(obj, asUnstructured, nil); err != nil {
		return obj, err
	}

	mapper := restmapper.NewDeferredDiscoveryRESTMapper(memory.NewMemCacheClient(kclient.Discovery()))
	mapping, err := mapper.RESTMapping(schema.GroupKind{Group: gvk.Group, Kind: gvk.Kind}, gvk.Version)
	if err != nil {
		return obj, err
	}

	actualObj, err := dynamicClient.Resource(mapping.Resource).Namespace("default").Create(asUnstructured, metav1.CreateOptions{DryRun: []string{metav1.DryRunAll}})
	if err != nil {
		return obj, err
	}

	return actualObj, nil
}
